from __future__ import annotations

import json
import subprocess
from collections import defaultdict
from collections.abc import Sequence
from typing import TYPE_CHECKING

from silence_lint_error.silencing import ErrorRunningTool
from silence_lint_error.silencing import Violation

if TYPE_CHECKING:
    from typing import TypeAlias

    FileName: TypeAlias = str
    RuleName: TypeAlias = str


class Semgrep:
    name = 'semgrep'

    def find_violations(
        self, rule_name: RuleName, filenames: Sequence[FileName],
    ) -> dict[FileName, list[Violation]]:
        proc = subprocess.run(
            (
                'semgrep', 'scan',
                '--metrics=off', '--oss-only',
                '--json',
                *filenames,
            ),
            capture_output=True,
            text=True,
        )

        if proc.returncode:
            raise ErrorRunningTool(proc)

        # extract filenames and line numbers
        results: dict[FileName, list[Violation]] = defaultdict(list)
        data = json.loads(proc.stdout)
        for result in data['results']:
            if result['check_id'] != rule_name:
                continue

            results[result['path']].append(
                Violation(
                    rule_name=result['check_id'],
                    lineno=result['start']['line'],
                ),
            )

        return dict(results)

    def silence_violations(
        self, src: str, violations: Sequence[Violation],
    ) -> str:
        [rule_name] = {violation.rule_name for violation in violations}
        linenos_to_silence = {violation.lineno for violation in violations}

        lines = src.splitlines(keepends=True)

        new_lines = []
        for current_lineno, line in enumerate(lines, start=1):
            if current_lineno in linenos_to_silence:
                leading_ws = line.removesuffix(line.lstrip())
                new_lines.append(f'{leading_ws}# nosemgrep: {rule_name}\n')
            new_lines.append(line)

        return ''.join(new_lines)



from __future__ import annotations

import argparse
import sys
from collections.abc import Sequence
from typing import NamedTuple

from silence_lint_error.fixing import Fixer
from silence_lint_error.fixing import Linter
from silence_lint_error.linters import fixit
from silence_lint_error.linters import ruff


LINTERS: dict[str, type[Linter]] = {
    'fixit': fixit.Fixit,
    'ruff': ruff.Ruff,
}


class Context(NamedTuple):
    rule_name: str
    file_names: list[str]
    linter: Linter


def _parse_args(argv: Sequence[str] | None) -> Context:
    parser = argparse.ArgumentParser(
        description=(
            'Fix linting errors by removing ignore/fixme comments '
            'and running auto-fixes.'
        ),
    )
    parser.add_argument(
        'linter', choices=LINTERS,
        help='The linter to use to fix the errors',
    )
    parser.add_argument('rule_name')
    parser.add_argument('filenames', nargs='*')
    args = parser.parse_args(argv)

    return Context(
        rule_name=args.rule_name,
        file_names=args.filenames,
        linter=LINTERS[args.linter](),
    )


def main(argv: Sequence[str] | None = None) -> int:
    rule_name, file_names, linter = _parse_args(argv)
    fixer = Fixer(linter)

    print('-> removing comments that silence errors', file=sys.stderr)
    changed_files = []
    for filename in file_names:
        try:
            fixer.unsilence_violations(rule_name=rule_name, filename=filename)
        except fixer.NoChangesMade:
            continue
        else:
            print(filename)
            changed_files.append(filename)

    if not changed_files:
        print('no silenced errors found', file=sys.stderr)
        return 0

    print(f'-> applying auto-fixes with {linter.name}', file=sys.stderr)
    ret, message = fixer.apply_fixes(rule_name=rule_name, filenames=changed_files)
    print(message, file=sys.stderr)

    return ret


if __name__ == '__main__':
    raise SystemExit(main())



from __future__ import annotations

import subprocess
from collections.abc import Sequence
from typing import Protocol

import attrs


@attrs.frozen
class Violation:
    rule_name: str
    lineno: int


@attrs.frozen
class ErrorRunningTool(Exception):
    proc: subprocess.CompletedProcess[str]


class Linter(Protocol):
    name: str

    def find_violations(
        self, rule_name: str, filenames: Sequence[str],
    ) -> dict[str, list[Violation]]:
        """Find violations of a rule.

        Returns:
            Mapping of file path to the violations found in that file.

        Raises:
            ErrorRunningTool: There was an error whilst running the linter.
        """

    def silence_violations(
        self, src: str, violations: Sequence[Violation],
    ) -> str:
        """Modify module source to silence violations.

        Returns:
            Modified `src` with comments that silence the `violations`.
        """


@attrs.frozen
class Silencer:
    linter: Linter

    class NoViolationsFound(Exception):
        pass

    @attrs.frozen
    class MultipleRulesViolated(Exception):
        rule_names: set[str]

    def find_violations(
            self, *, rule_name: str, file_names: Sequence[str],
    ) -> dict[str, list[Violation]]:
        violations = self.linter.find_violations(rule_name, file_names)

        if not violations:
            raise self.NoViolationsFound

        violation_names = {
            violation.rule_name
            for file_violations in violations.values()
            for violation in file_violations
        }
        if len(violation_names) != 1:
            raise self.MultipleRulesViolated(violation_names)

        return violations

    def silence_violations(
            self, *, filename: str, violations: Sequence[Violation],
    ) -> bool:
        with open(filename) as f:
            src = f.read()

        src_with_comments = self.linter.silence_violations(src, violations)

        with open(filename, 'w') as f:
            f.write(src_with_comments)

        return src_with_comments != src



from __future__ import annotations

import subprocess
from collections import defaultdict
from collections.abc import Sequence
from typing import TYPE_CHECKING

from silence_lint_error import comments
from silence_lint_error.silencing import ErrorRunningTool
from silence_lint_error.silencing import Violation

if TYPE_CHECKING:
    from typing import TypeAlias

    FileName: TypeAlias = str
    RuleName: TypeAlias = str


class Flake8:
    name = 'flake8'

    def find_violations(
        self, rule_name: RuleName, filenames: Sequence[FileName],
    ) -> dict[FileName, list[Violation]]:
        proc = subprocess.run(
            (
                'flake8',
                '--select', rule_name,
                '--format', '%(path)s %(row)s',
                *filenames,
            ),
            capture_output=True,
            text=True,
        )

        if proc.returncode and proc.stderr.endswith('No module named flake8\n'):
            raise ErrorRunningTool(proc)

        # extract filenames and line numbers
        results: dict[FileName, list[Violation]] = defaultdict(list)
        for line in proc.stdout.splitlines():
            filename_, lineno_ = line.rsplit(maxsplit=1)
            results[filename_].append(Violation(rule_name, int(lineno_)))

        return results

    def silence_violations(
        self, src: str, violations: Sequence[Violation],
    ) -> str:
        [rule_name] = {violation.rule_name for violation in violations}
        linenos_to_silence = {violation.lineno for violation in violations}
        return comments.add_noqa_comments(src, linenos_to_silence, rule_name)



from __future__ import annotations

import re
import subprocess
from collections import defaultdict
from collections.abc import Iterator
from collections.abc import Sequence
from typing import TYPE_CHECKING

from silence_lint_error import comments
from silence_lint_error.silencing import ErrorRunningTool
from silence_lint_error.silencing import Violation

if TYPE_CHECKING:
    from typing import TypeAlias

    FileName: TypeAlias = str
    RuleName: TypeAlias = str


class Fixit:
    name = 'fixit'

    def __init__(self) -> None:
        self.error_line_re = re.compile(r'^.*?@\d+:\d+ ')

    def find_violations(
        self, rule_name: RuleName, filenames: Sequence[FileName],
    ) -> dict[FileName, list[Violation]]:
        proc = subprocess.run(
            (
                'fixit',
                '--rules', rule_name,
                'lint', *filenames,
            ),
            capture_output=True,
            text=True,
        )

        if proc.returncode and proc.stderr.endswith('No module named fixit\n'):
            raise ErrorRunningTool(proc)

        # extract filenames and line numbers
        results: dict[str, list[Violation]] = defaultdict(list)
        for line in proc.stdout.splitlines():
            found_error = self._parse_output_line(line)
            if found_error:
                filename, violation = found_error
                results[filename].append(violation)
            else:  # pragma: no cover
                pass

        return results

    def _parse_output_line(
            self, line: str,
    ) -> tuple[FileName, Violation] | None:
        if not self.error_line_re.match(line):
            return None

        location, violated_rule_name, *__ = line.split(maxsplit=2)
        filename, position = location.split('@', maxsplit=1)
        lineno, *__ = position.split(':', maxsplit=1)

        rule_name_ = violated_rule_name.removesuffix(':')
        return filename, Violation(rule_name_, int(lineno))

    def silence_violations(
        self, src: str, violations: Sequence[Violation],
    ) -> str:
        [rule_name] = {violation.rule_name for violation in violations}
        linenos_to_silence = {violation.lineno for violation in violations}

        lines = src.splitlines(keepends=True)

        new_lines = []
        for current_lineno, line in enumerate(lines, start=1):
            if current_lineno in linenos_to_silence:
                leading_ws = line.removesuffix(line.lstrip())
                new_lines.append(f'{leading_ws}# lint-fixme: {rule_name}\n')
            new_lines.append(line)

        return ''.join(new_lines)

    def remove_silence_comments(self, src: str, rule_name: RuleName) -> str:
        return ''.join(
            self._remove_comments(
                src.splitlines(keepends=True), rule_name,
            ),
        )

    def _remove_comments(
            self, lines: Sequence[str], rule_name: RuleName,
    ) -> Iterator[str]:
        __, rule_id = rule_name.rsplit(':', maxsplit=1)
        fixme_comment = f'# lint-fixme: {rule_id}'
        for line in lines:
            if line.strip() == fixme_comment:  # fixme comment only
                continue
            elif line.rstrip().endswith(fixme_comment):  # code then fixme comment
                trailing_ws = line.removeprefix(line.rstrip())
                line_without_comment = (
                    line.rstrip().removesuffix(fixme_comment)  # remove comment
                    .rstrip()  # and remove any intermediate ws
                )
                yield line_without_comment + trailing_ws
            else:
                yield line

    def apply_fixes(
            self, rule_name: RuleName, filenames: Sequence[str],
    ) -> tuple[int, str]:
        proc = subprocess.run(
            (
                'fixit',
                '--rules', rule_name,
                'fix', '--automatic', *filenames,
            ),
            capture_output=True, text=True,
        )
        return proc.returncode, proc.stderr.strip()


class FixitInline(Fixit):
    """An alternative `fixit` implementation that adds `lint-fixme` comment inline.

    This is sometimes necessary because `fixit` does not always respect `lint-fixme`
    comments when they are on the line above the line causing the error. This is a
    known bug and is reported in https://github.com/Instagram/Fixit/issues/405.

    In some of these cases, placing the comment on the same line as the error can
    ensure it is respected (e.g. for decorators).
    """

    def silence_violations(
        self, src: str, violations: Sequence[Violation],
    ) -> str:
        [rule_name] = {violation.rule_name for violation in violations}
        linenos_to_silence = {violation.lineno for violation in violations}
        return comments.add_error_silencing_comments(
            src, linenos_to_silence,
            'lint-fixme', rule_name,
        )



from __future__ import annotations

import tokenize_rt


def add_error_silencing_comments(
        src: str, error_lines: set[int],
        comment_type: str, error_code: str,
) -> str:
    """Add comments to some code to silence linting errors.

    Args:
        src: The content of the module to add comments to.
        error_lines: The lines on which to silence errors.
        comment_type: The type of comment to add (e.g. `noqa` or `lint-fixme`)
        code: The error code to silence.

    Returns:
        The content of the module with the additional comments added.
    """
    tokens = tokenize_rt.src_to_tokens(src)

    for idx, token in tokenize_rt.reversed_enumerate(tokens):
        if token.line not in error_lines:
            continue
        if not token.src.strip():
            continue

        if token.name == 'COMMENT':
            new_comment = add_code_to_comment(token.src, comment_type, error_code)
            tokens[idx] = tokens[idx]._replace(src=new_comment)
        else:
            tokens.insert(
                idx+1, tokenize_rt.Token(
                    'COMMENT', f'# {comment_type}: {error_code}',
                ),
            )
            tokens.insert(idx+1, tokenize_rt.Token('UNIMPORTANT_WS', '  '))

        error_lines.remove(token.line)

    return tokenize_rt.tokens_to_src(tokens)


def add_noqa_comments(src: str, lines: set[int], error_code: str) -> str:
    """Add `noqa` comments to some code.

    Args:
        src: The content of the module to add `noqa` comments to.
        lines: The lines on which to add `noqa` coments.
        code: The error code to silence.

    Returns:
        The content of the module with the additional comments added.
    """
    return add_error_silencing_comments(src, lines, 'noqa', error_code)


def add_code_to_comment(comment: str, comment_type: str, code: str) -> str:
    """Add to a comment to make it a error-silencing comment.

    If the comment already includes an error silencing section of the same type, this
    will add the code to the list of silenced errors.
    """
    if f'{comment_type}: ' in comment:
        return comment.replace(
            f'{comment_type}: ', f'{comment_type}: {code},', 1,
        )
    else:
        return comment + f'  # {comment_type}: {code}'


def remove_error_silencing_comments(
        src: str,
        comment_type: str, error_code: str,
) -> str:
    """Remove comments that silence linting errors from some code.

    Args:
        src: The content of the module to remove comments from.
        comment_type: The type of comment to remove (e.g. `noqa` or `lint-fixme`)
        error_code: The error code that is silenced.

    Returns:
        The content of the module without the comments that slence this error code.
    """
    tokens = tokenize_rt.src_to_tokens(src)

    for idx, token in tokenize_rt.reversed_enumerate(tokens):
        if (
                token.name == 'COMMENT'
                and comment_type in token.src and error_code in token.src
        ):
            new_comment = remove_code_from_comment(
                token.src, comment_type, error_code,
            )
            if new_comment:
                tokens[idx] = tokens[idx]._replace(src=new_comment)
            else:
                tokens.pop(idx)

        # delete trailing whitespace caused by removing comments
        elif (
                token.name == 'UNIMPORTANT_WS'
                and tokens[idx+1].name in {'NEWLINE', 'NL'}
        ):
            tokens.pop(idx)

    return tokenize_rt.tokens_to_src(tokens)


def remove_code_from_comment(comment: str, comment_type: str, code: str) -> str:
    """Remove the error-silencing portion from a comment."""
    return (
        comment
        .replace(f'{comment_type}: {code},', f'{comment_type}: ')
        .replace(f',{code}', '')
        .removeprefix(f'# {comment_type}: {code}')
        .removesuffix(f'{comment_type}: {code}')
        .strip()
        .removesuffix('  #')
    )



from __future__ import annotations

import json
import subprocess
from collections import defaultdict
from collections.abc import Sequence
from typing import TYPE_CHECKING

from silence_lint_error import comments
from silence_lint_error.silencing import ErrorRunningTool
from silence_lint_error.silencing import Violation

if TYPE_CHECKING:
    from typing import TypeAlias

    FileName: TypeAlias = str
    RuleName: TypeAlias = str


class Ruff:
    name = 'ruff'

    def find_violations(
        self, rule_name: RuleName, filenames: Sequence[FileName],
    ) -> dict[FileName, list[Violation]]:
        proc = subprocess.run(
            (
                'ruff', 'check',
                '--select', rule_name,
                '--output-format', 'json',
                *filenames,
            ),
            capture_output=True,
            text=True,
        )

        if proc.returncode and proc.stderr.endswith('No module named ruff\n'):
            raise ErrorRunningTool(proc)

        # extract filenames and line numbers
        all_violations = json.loads(proc.stdout)
        results: dict[FileName, list[Violation]] = defaultdict(list)
        for violation in all_violations:
            if violation['code'] is None:
                # ignore syntax errors while parsing the file
                continue

            results[violation['filename']].append(
                Violation(
                    rule_name=violation['code'],
                    lineno=violation['location']['row'],
                ),
            )

        return results

    def silence_violations(
        self, src: str, violations: Sequence[Violation],
    ) -> str:
        [rule_name] = {violation.rule_name for violation in violations}
        linenos_to_silence = {violation.lineno for violation in violations}
        return comments.add_noqa_comments(src, linenos_to_silence, rule_name)

    def remove_silence_comments(self, src: str, rule_name: RuleName) -> str:
        return comments.remove_error_silencing_comments(
            src, comment_type='noqa', error_code=rule_name,
        )

    def apply_fixes(
            self, rule_name: RuleName, filenames: Sequence[str],
    ) -> tuple[int, str]:
        proc = subprocess.run(
            (
                'ruff', 'check',
                '--fix',
                '--select', rule_name,
                *filenames,
            ),
            capture_output=True, text=True,
        )
        return proc.returncode, proc.stdout.strip()



from __future__ import annotations

import subprocess
from collections import defaultdict
from collections.abc import Sequence
from typing import TYPE_CHECKING

import tokenize_rt

from silence_lint_error.silencing import ErrorRunningTool
from silence_lint_error.silencing import Violation

if TYPE_CHECKING:
    from typing import TypeAlias

    FileName: TypeAlias = str
    RuleName: TypeAlias = str


class Mypy:
    name = 'mypy'

    def find_violations(
        self, rule_name: RuleName, filenames: Sequence[FileName],
    ) -> dict[FileName, list[Violation]]:
        proc = subprocess.run(
            (
                'mypy',
                '--follow-imports', 'silent',  # do not report errors in other modules
                '--enable-error-code', rule_name,
                '--show-error-codes', '--no-pretty', '--no-error-summary',
                *filenames,
            ),
            capture_output=True,
            text=True,
        )

        if proc.returncode > 1:
            raise ErrorRunningTool(proc)

        # extract filenames and line numbers
        results: dict[FileName, list[Violation]] = defaultdict(list)
        for line in proc.stdout.splitlines():
            if not line.endswith(f'[{rule_name}]'):
                continue

            location, *__ = line.split()
            filename_, lineno_, *__ = location.split(':')

            results[filename_].append(Violation(rule_name, int(lineno_)))

        return results

    def silence_violations(
        self, src: str, violations: Sequence[Violation],
    ) -> str:
        rule_name = violations[0].rule_name
        lines_with_errors = {
            violation.lineno
            for violation in violations
        }

        tokens = tokenize_rt.src_to_tokens(src)
        for idx, token in tokenize_rt.reversed_enumerate(tokens):
            if token.line not in lines_with_errors:
                continue
            if not token.src.strip():
                continue

            if token.name == 'COMMENT':
                if 'type: ignore' in token.src:
                    prefix, __, ignored = token.src.partition('type: ignore')
                    codes = ignored.strip('[]').split(',')
                    codes += [rule_name]
                    new_comment = f'{prefix}type: ignore[{",".join(codes)}]'
                else:
                    new_comment = token.src + f'  # type: ignore[{rule_name}]'
                tokens[idx] = tokens[idx]._replace(src=new_comment)
            else:
                tokens.insert(
                    idx+1, tokenize_rt.Token(
                        'COMMENT', f'# type: ignore[{rule_name}]',
                    ),
                )
                tokens.insert(idx+1, tokenize_rt.Token('UNIMPORTANT_WS', '  '))

            lines_with_errors.remove(token.line)

        return tokenize_rt.tokens_to_src(tokens)



from __future__ import annotations

import argparse
import sys
from collections.abc import Sequence
from typing import NamedTuple

from silence_lint_error.linters import fixit
from silence_lint_error.linters import flake8
from silence_lint_error.linters import mypy
from silence_lint_error.linters import ruff
from silence_lint_error.linters import semgrep
from silence_lint_error.silencing import ErrorRunningTool
from silence_lint_error.silencing import Linter
from silence_lint_error.silencing import Silencer


LINTERS: dict[str, type[Linter]] = {
    'fixit': fixit.Fixit,
    'fixit-inline': fixit.FixitInline,
    'flake8': flake8.Flake8,
    'mypy': mypy.Mypy,
    'ruff': ruff.Ruff,
    'semgrep': semgrep.Semgrep,
}


class Context(NamedTuple):
    rule_name: str
    file_names: list[str]
    linter: Linter


def _parse_args(argv: Sequence[str] | None) -> Context:
    parser = argparse.ArgumentParser(
        description='Ignore linting errors by adding ignore/fixme comments.',
    )
    parser.add_argument(
        'linter', choices=LINTERS,
        help='The linter for which to ignore errors',
    )
    parser.add_argument('rule_name')
    parser.add_argument('filenames', nargs='*')
    args = parser.parse_args(argv)

    return Context(
        rule_name=args.rule_name,
        file_names=args.filenames,
        linter=LINTERS[args.linter](),
    )


def main(argv: Sequence[str] | None = None) -> int:
    rule_name, file_names, linter = _parse_args(argv)
    silencer = Silencer(linter)

    print(f'-> finding errors with {linter.name}', file=sys.stderr)
    try:
        violations = silencer.find_violations(
            rule_name=rule_name, file_names=file_names,
        )
    except ErrorRunningTool as e:
        print(f'ERROR: {e.proc.stderr.strip()}', file=sys.stderr)
        return e.proc.returncode
    except silencer.NoViolationsFound:
        print('no errors found', file=sys.stderr)
        return 0
    except silencer.MultipleRulesViolated as e:
        print(
            'ERROR: errors found for multiple rules:', sorted(e.rule_names),
            file=sys.stderr,
        )
        return 1
    else:
        print(f'found errors in {len(violations)} files', file=sys.stderr)

    print('-> adding comments to silence errors', file=sys.stderr)
    ret = 0
    for filename, file_violations in violations.items():
        print(filename)
        ret |= silencer.silence_violations(
            filename=filename, violations=file_violations,
        )

    return ret


if __name__ == '__main__':
    raise SystemExit(main())



from __future__ import annotations

from collections.abc import Sequence
from typing import Protocol

import attrs


class Linter(Protocol):
    name: str

    def remove_silence_comments(self, src: str, rule_name: str) -> str:
        """Remove comments that silence rule violations.

        Returns:
            Modified `src` without comments that silence the `violations`.
        """

    def apply_fixes(
            self, rule_name: str, filenames: Sequence[str],
    ) -> tuple[int, str]:
        """Fix violations of a rule.

        Returns:
            Return code and stdout from the process that fixed the violations.
        """


@attrs.frozen
class Fixer:
    linter: Linter

    class NoChangesMade(Exception):
        pass

    def unsilence_violations(
            self, *, rule_name: str, filename: str,
    ) -> None:
        with open(filename) as f:
            src = f.read()

        src_without_comments = self.linter.remove_silence_comments(src, rule_name)

        if src_without_comments == src:
            raise self.NoChangesMade

        with open(filename, 'w') as f:
            f.write(src_without_comments)

    def apply_fixes(
            self, *, rule_name: str, filenames: Sequence[str],
    ) -> tuple[int, str]:
        return self.linter.apply_fixes(rule_name, filenames)
